"""
plot module handles math plotting operations
"""

from termgraph import termgraph as tg

# print("# " + "Images Distribution Over Classes")
#
# tg.print_categories(['t1'] * 10, list(range(90, 100)))
#
# tg.chart(
#     # len_categories=2,
#     colors=[92, 96],
#     labels=['train', 'test'],
#     data=[[5000, 1000], [5000, 1000]],
#     args=dict(
#         width=50,
#         suffix=" images",
#         format="{:<5.2f}",
#         stacked=False,
#         vertical=False,
#         no_labels=False,
#         verbose=True,
#         version=True,
#         title='TEST CHART ssssssss',
#         verbos=True,
#         histogram=False,
#         no_values=False,
#         different_scale=False,
#     ),
#     # normal_data=[[50, 10], [50, 10]],/
# )
cli_chart_color_map = {'black': 90, 'red': 91, 'green': 92, 'yellow': 93, 'blue': 94, 'pink': 95, 'cyan': 96,
                       'white': 97, 'grey': 99}


def plt_cmd_bar(name, labels, data, y_categories, colors, y_label):
    print("# " + name)
    color_map_list = [cli_chart_color_map[x] for x in colors]
    tg.print_categories(y_categories, color_map_list)
    suffix = " " + y_label

    tg.chart(
        colors=color_map_list,
        labels=labels,
        data=data,
        args=dict(
            width=50,
            suffix=suffix,
            format="{:<5.2f}",
            stacked=False,
            vertical=False,
            no_labels=False,
            verbose=True,
            version=True,
            verbos=True,
            histogram=False,
            no_values=False,
            different_scale=False,
        ),
    )



import re
import shutil
from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple

from boto3 import Session
from commons.exceptions import CLIException, ExistCode

from dsdlsdk.storage import LocalDiskStorage, S3Storage, SftpStorage, Storage


class StorageBuilder(object):

    @staticmethod
    def build_by_name(storage_name: str, config: Dict) -> Storage:
        """
        Build storage client by name
        Args:
            storage_name:
            config:

        Returns:

        """
        if "storage" in config.keys():
            storage_config = config['storage'].get(storage_name, None)
            if storage_config is not None:
                path = storage_config.get('path', None)
                if path is not None:
                    if path.startswith('s3://'):
                        endpoint = storage_config.get('endpoint', None)
                        access_key = storage_config.get('ak', None)
                        secret_key = storage_config.get('sk', None)
                        return S3Storage(path, endpoint, access_key,
                                         secret_key)
                    elif path.startswith('sftp://'):
                        return SftpStorage()
                    else:
                        return LocalDiskStorage()
                else:
                    raise CLIException(ExistCode.NOT_SUPPORTED_STORE_PATH,
                                       f'unsupported storage path: {path}')
            else:
                raise CLIException(
                    ExistCode.STORAGE_NAME_NOT_EXIST,
                    f"storage name key `{storage_name}' not found in config")
        else:
            raise CLIException(
                ExistCode.STORAGE_NOT_EXIST,
                f"storage named `{storage_name}' not found in config")


if __name__ == "__main__":
    s: Storage = StorageBuilder.build_by_name(
        "s3", {
            "storage": {
                "s3": {
                    "endpoint": "http://10.140.0.94:9800",
                    "access_key": "ailabminio",
                    "secret_key": "123123123",
                    "path": "s3://testdata"
                }
            }
        })
    b = s.remove_tree("s3://testdata/.dsdl")
    print(b)



from io import BytesIO
from typing import Dict, Optional, Tuple

import numpy as np
import pandas as pd
import requests
import streamlit as st
from PIL import Image

from src.types import Array, Color, PIL_Image
from src.utils.draw_bbox import draw_bboxes

HOUR = 60 * 60


@st.experimental_memo(max_entries=300, ttl=HOUR)
def open_image(url: str) -> PIL_Image:
    response = requests.get(url)
    return Image.open(BytesIO(response.content)).convert("RGB")


def resize_with_pad(
    pil_img: PIL_Image, shape: Tuple[int, int], background_color: Color = (0, 0, 0)
) -> PIL_Image:
    width, height = pil_img.size
    final_width, final_height = shape
    ratio = min(final_width / width, final_height / height)
    result = Image.new(pil_img.mode, shape, background_color)
    pil_img = pil_img.resize((int(ratio * width), int(ratio * height)))
    width, height = pil_img.size
    result.paste(pil_img, ((final_width - width) // 2, (final_height - height) // 2))
    return result


@st.experimental_memo(max_entries=100, ttl=0.1 * HOUR)
def load_crop(
    url: str,
    x1y1x2y2: Tuple[int, int, int, int],
    shape: Optional[Tuple[int, int]] = None,
) -> Array:
    image = open_image(url)
    crop = image.crop(x1y1x2y2)
    if shape is not None:
        crop = resize_with_pad(crop, shape=shape)
    return np.array(crop)


@st.experimental_memo(max_entries=100, ttl=0.1 * HOUR)
def load_and_annotate_image(
    url: str, image_annotations: pd.Series, color_map: Dict[str, Color]
) -> Array:
    image = np.array(open_image(url))
    boxes = image_annotations[["x1", "y1", "x2", "y2"]].values
    box_colors = [
        color_map[category_name] for category_name in image_annotations["category_name"]
    ]
    image = draw_bboxes(image, boxes, box_colors, image_annotations["category_name"])
    return image



import argparse
import os
import sys
from argparse import Action, HelpFormatter


class EnvDefaultVar(Action):
    """
    默认参数使用环境变量。
    优先级为：
    显式设置的值 > default值 > 环境变量值
    """

    def __init__(self, envvar, required=False, default=None, **kwargs):
        if not default and envvar:  # 没有默认值，但是给了环境变量key
            if envvar in os.environ:
                default = os.environ[envvar]
        if required and default:  # 一旦得到了默认值，那么改变required的值为False, 这样用户不必一定要在命令行显式敲出来--dataset-name这个选项
            required = False
        super(EnvDefaultVar, self).__init__(default=default, required=required, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)


class DsdlArgumentParser(argparse.ArgumentParser):
    """
    重写ArgumentParser的format_usage方法，使得usage信息可以自动换行。
    """

    def __init__(self, example='',  *args, **kwargs):
        self.__example = example
        super(DsdlArgumentParser, self).__init__(*args, **kwargs)

    def format_usage(self):
        if self.usage:
            return f"usage: {self.usage}"
        else:
            formatter = self._get_formatter()
            formatter.add_usage(self.usage, self._actions,
                                self._mutually_exclusive_groups)
            return formatter.format_help()

    def add_subparsers(self, **kwargs):
        """
        重写add_subparsers方法，使得help信息可以自动换行。
        """
        kwargs.setdefault('prog', self.prog)
        return super(DsdlArgumentParser, self).add_subparsers(**kwargs)

    def add_example(self, formatter):
        # example
        if self.__example:
            # read txt from resources/example.txt
            example_file = os.path.join(os.path.dirname(__file__), '../resources', self.__example)
            if os.path.exists(example_file):
                with open(example_file, 'r') as f:
                    example_str = f.read()
                    formatter.start_section("Examples")
                    formatter.add_raw_text(example_str)
                    formatter.end_section()

    def format_help(self) -> str:
        formatter = self._get_formatter()

        # usage
        formatter.add_usage(self.usage, self._actions,
                            self._mutually_exclusive_groups)

        # description
        formatter.add_text(self.description)

        # positionals, optionals and user-defined groups
        for action_group in self._action_groups:
            formatter.start_section(action_group.title)
            formatter.add_text(action_group.description)
            formatter.add_arguments(action_group._group_actions)
            formatter.end_section()

        # example
        self.add_example(formatter)

        # epilog
        formatter.add_text(self.epilog)

        # determine help from format above
        return formatter.format_help()


class CustomHelpFormatter(HelpFormatter):
    """
    自定义命令输出格式
    """

    def _format_action(self, action):
        if type(action) == argparse._SubParsersAction:
            # inject new class variable for subcommand formatting
            subactions = action._get_subactions()
            invocations = [self._format_action_invocation(a) for a in subactions]
            self._subcommand_max_length = max(len(i) for i in invocations)

        if type(action) == argparse._SubParsersAction._ChoicesPseudoAction:
            # format subcommand help line
            subcommand = self._format_action_invocation(action)
            width = self._subcommand_max_length
            help_text = ""
            if action.help:
                help_text = self._expand_help(action)
            return "  {:{width}} -  {}\n".format(subcommand, help_text, width=width)

        elif type(action) == argparse._SubParsersAction:
            # process subcommand help section
            msg = '\n'
            for subaction in action._get_subactions():
                msg += self._format_action(subaction)
            return msg
        else:
            return super(CustomHelpFormatter, self)._format_action(action)

    def format_help(self):
        return super().format_help()

    def add_raw_text(self, text):
        if text is not None:
            if '%(prog)' in text:
                text = text % dict(prog=self._prog)
            self._add_item(lambda t : t, [text])



"""
一个命令的实现例子

Examples:
    $python cli.py rm  -y  openmmlab/flying3d

"""

from commands.cmdbase import CmdBase
from commons.stdio import print_stdout


class Search(CmdBase):

    def __init__(self):
        self.__config = None

    def init_parser(self, subparsers):
        """
        Initialize the parser for the command
        document : https://docs.python.org/zh-cn/3/library/argparse.html#

        Args:
            subparsers:

        Returns:

        """
        search_parser = subparsers.add_parser(
            'search',
            help='search dataset in a desired manner',
        )  # example 样例文件位于resources/下，普通的文本文件，每个命令写一个

        return search_parser

    def cmd_entry(self, cmdargs, config, *args, **kwargs):
        """
        Entry point for the command

        Args:
            self:
            cmdargs:
            config:

        Returns:

        """
        print_stdout("""
CIFAR-10-Auto     21M
CIFAR-10          241M
CIFAR-100         243M
Fashion-MNIST     281M
STL-10            2.1G""")



from importlib.metadata import version, PackageNotFoundError
# import git
from setuptools_scm import get_version

try:
    __short_version__ = version("odl-cli") # return '0.1.1'
except PackageNotFoundError:
    # package is not installed
    pass

# __full_version__ = get_version() # return '0.1.dev519+g37fcc3b.d20221206'

# __version_tuple__ = tuple(__full_version__.split("."))


# __version_dev_part__ = ''
# for ele in __version_tuple__:    
#     if ele.startswith('dev'):
#         _tmp_tuple = tuple(ele.split("+"))
#         for ele2 in _tmp_tuple:
#             if ele2.startswith('dev'):
#                 __version_dev_part__ = ele2 # return 'devxxx'
#                 break
    
# repo = git.Repo(search_parent_directories=True)
# sha = repo.head.object.hexsha
# #just keep the first 4 characters
# __version_commitID_part__ = sha.split()[0][:4]

# __version__ = 'v' + __short_version__  + '.' + __version_dev_part__ + __version_commitID_part__
# __version__ = 'v' + __short_version__  + '.' + __version_dev_part__
__version__ = 'v' + __short_version__


"""

Examples:

class Example(CmdBase):
    def init_parser(self, subparsers):
        status_parser = subparsers.add_parser('example', help='Show the working tree status')
        status_parser.add_argument('show', nargs='?', default='SHOW', help='show example')
        return status_parser

    def cmd_entry(self, args, config):
        print(args)
        print(config)

"""
from abc import ABC, abstractmethod
from argparse import ArgumentParser, _SubParsersAction

from commons.exceptions import CLIException
from commons.stdio import print_stderr
from loguru import logger


class CmdBase(ABC):
    """
    Base class for all commands

    Args:
        ABC:

    Returns:

    """

    def __init__(self):
        self._parser: ArgumentParser = None

    def cmd_main(self, cmdargs, config, *args, **kwargs):
        try:
            self.cmd_entry(cmdargs, config, *args, **kwargs)
        except CLIException as e:
            logger.exception(e.message)
            print_stderr(e.message)
            exit(e.errcode)
        except Exception as e:
            # TODO log
            logger.exception(e)
            print_stderr("unknown error, please see log files at ~/.dsdl/logs")
            exit(-1)
        exit(0)

    def setup_parser(self, subparsers: _SubParsersAction) -> ArgumentParser:
        self._parser = self.init_parser(subparsers)
        return self._parser

    @abstractmethod
    def cmd_entry(self, cmdargs, config, *args, **kwargs):
        """
        Entry point for the command

        Args:
            self:
            args:
            config:

        Returns:

        """
        raise NotImplementedError

    @abstractmethod
    def init_parser(self, subparsers: _SubParsersAction) -> ArgumentParser:
        """
        Initialize the parser for the command

        Args:
            self:
            subparsers:

        Returns:

        """
        raise NotImplementedError



import json

from commons.stdio import print_stderr, print_stdout
from loguru import logger
from rich import print as rprint
from rich.pretty import pprint

from .cmdbase import CmdBase
from .const import (DEFAULT_CLI_CONFIG_FILE, DEFAULT_CONFIG_DIR,
                    DEFAULT_LOCAL_STORAGE_PATH, PROG_NAME, SQLITE_DB_PATH)


class Config(CmdBase):
    """Set dsdl configuration {file path, user login info}.

    Args:
        CmdBase (_type_): _description_
    """

    def init_parser(self, subparsers):
        """_summary_

        _extended_summary_

        Args:
            subparsers (_type_): _description_
        """

        config_parser = subparsers.add_parser(
            'config',
            help='set odl-cli configuration.',
            example='config.example')
        config_parser.add_argument('-k',
                                   '--keys',
                                   action='store_true',
                                   help='show all the available keys',
                                   required=False)
        config_parser.add_argument('-l',
                                   '--list',
                                   action='count',
                                   help='show all key value pairs',
                                   required=False)

        sub_config_parser = config_parser.add_subparsers(dest='command')
        repo_parser = sub_config_parser.add_parser(
            'repo', help='set repo configuration')
        repo_parser.add_argument('--repo-name',
                                 help='set repo name',
                                 required=True)
        repo_parser.add_argument('--repo-username', help='set repo user name')
        repo_parser.add_argument('--repo-userpswd',
                                 help='set repo user password')
        repo_parser.add_argument('--repo-service', help='set repo service url')
        repo_parser.add_argument('--repo-remove',
                                 action='store_true',
                                 help='remove specific configuration',
                                 required=False)

        storage_parser = sub_config_parser.add_parser(
            'storage',
            help='set storage configuration, only support local/s3/sftp for now'
        )
        storage_parser.add_argument('--storage-name',
                                    help='set storage name',
                                    required=True)
        storage_parser.add_argument('--storage-path',
                                    help='set storage path',
                                    required=True)
        storage_parser.add_argument('--storage-credentials',
                                    action='append',
                                    nargs=2,
                                    help='set credentials',
                                    required=False)
        storage_parser.add_argument('--storage-endpoint',
                                    help='set storage endpoint',
                                    default='',
                                    required=False)
        storage_parser.add_argument(
            '--storage-remove',
            action='store_true',
            help='remove specific storage configuration',
            required=False)
        return config_parser

    @logger.catch
    def cmd_entry(self, args, config):
        """
        Entry point for the command.

        Args:
            self:
            args:
            config:

        Returns:

        """
        # if not args.repo_name:
        #             logger.exception('Please name a repo using {} before you can set is info'.format('--repo-name'))
        if args.command:
            # repo command handler
            if args.command == 'repo':
                if args.repo_remove:
                    try:
                        self.__repo_delete(config, args)
                        self.__config_writter(config)
                        print_stdout('REPO config for {} was removed !'.format(
                            args.repo_name))
                        logger.info('REPO config for {} was removed !'.format(
                            args.repo_name))
                    except KeyError as err:
                        print_stdout('No repo named {}!'.format(
                            args.repo_name))
                        logger.exception('No repo named {}'.format(
                            args.repo_name))
                elif (not args.repo_remove) & (args.repo_name
                                               not in config['repo'].keys()):
                    # new repo
                    self.__repo_new(config, args)
                    print_stdout('Your repo config for {} success !'.format(
                        args.repo_name))
                    logger.info('REPO: {} config success !'.format(
                        args.repo_name))
                    self.__config_writter(config)
                else:
                    print_stdout(
                        'REPO config for {} already exists, please remove and re-configure it !'
                        .format(args.repo_name))
                    logger.error(
                        'REPO config for {} already exists, please remove and re-configure it !'
                        .format(args.repo_name))

            elif args.command == 'storage':
                if args.storage_remove:
                    try:
                        self.__storage_delete(config, args)
                        self.__config_writter(config)
                        print_stdout(
                            'STORAGE config for {} was removed !'.format(
                                args.storage_name))
                        logger.info(
                            'STORAGE config for {} was removed !'.format(
                                args.storage_name))
                    except KeyError as err:
                        print_stdout('No storage named {} !'.format(
                            args.storage_name))
                        logger.exception('No storage named {} !'.format(
                            args.storage_name))

                # new storage entry
                elif (not args.storage_remove) & (
                        args.storage_name not in config['storage'].keys()):
                    self.__storage_new(config, args)
                    self.__config_writter(config)
                # old storage update
                else:
                    print_stdout(
                        'STORAGE config for {} already exists, please remove and re-configure it !'
                        .format(args.storage_name))
                    logger.error(
                        'STORAGE config for {} already exists, please remove and re-configure it !'
                        .format(args.storage_name))
        else:
            if args.keys:
                snippet_keys = """The available keys are:
                \nrepo-name           # set repo name\nrepo-username       # set repo username\nrepo-userpswd       # set repo password\nrepo-service        # set repo service url
                \nstorage-name        # storage name\nstorage-path        # storage path\nstorage-credentials # storage credentials S3:(access-key, secret-key), SFTP:(username,password)\nstorage-endpoint    # storage endpoint
                """
                rprint(snippet_keys)

            if args.list:
                if args.list == 1:
                    pprint(config)
                else:
                    pprint(config, expand_all=True)

            if not (args.keys or args.list):
                helper_str = """Please use -h to get more information. """
                # rprint(helper_str)
                self._parser.print_help()

    def __config_writter(self, config):
        with open(DEFAULT_CLI_CONFIG_FILE, 'w') as file:
            return json.dump(config, file, indent=4)

    def __repo_new(self, config, args):
        config['repo'][args.repo_name] = {}
        if args.repo_username:
            config['repo'][args.repo_name]['user'] = args.repo_username
        if args.repo_userpswd:
            config['repo'][args.repo_name]['passwd'] = args.repo_userpswd
        if args.repo_service:
            config['repo'][args.repo_name]['service'] = args.repo_service

    def __repo_delete(self, config, args):
        del config['repo'][args.repo_name]

    def __storage_new(self, config, args):
        config['storage'][args.storage_name] = {}

        if args.storage_path[:2] not in ['s3', 'sf']:
            config['storage'][args.storage_name]['path'] = args.storage_path
            print_stdout(
                'STORAGE LOCAL: The path {} configuration is successful!'
                .format(args.storage_path))
            logger.info(
                'STORAGE LOCAL: The path {} configuration is successful!'
                .format(args.storage_path))

        elif args.storage_path[:2] == 's3':
            config['storage'][
                args.storage_name]['ak'] = args.storage_credentials[0][0]
            config['storage'][
                args.storage_name]['sk'] = args.storage_credentials[0][1]
            config['storage'][args.storage_name]['path'] = args.storage_path
            config['storage'][
                args.storage_name]['endpoint'] = args.storage_endpoint
            print_stdout('STORAGE S3: {} config success !'.format(
                args.storage_name))
            logger.info('STORAGE S3: {} config success !'.format(
                args.storage_name))

        elif args.storage_path[:4] == 'sftp':
            config['storage'][
                args.storage_name]['user'] = args.storage_credentials[0][0]
            config['storage'][
                args.storage_name]['password'] = args.storage_credentials[0][1]
            config['storage'][args.storage_name]['path'] = args.storage_path
            print_stdout('STORAGE SFTP: {} config success !'.format(
                args.storage_name))
            logger.info('STORAGE STFP: {} config success !'.format(
                args.storage_name))

    def __storage_delete(self, config, args):
        del config['storage'][args.storage_name]



import os

DSDL_CLI_DATASET_NAME = "DATASET_NAME"  # cd命令设置的环境变量的数据集名称

DEFAULT_CONFIG_DIR = os.path.join(
    os.path.expanduser("~"), ".dsdl"
)  # 默认配置目录,放在用户的家目录下的.dsdl目录

__DEFAULT_CLI_CONFIG_FILE_NAME = "dsdl.json"  # 默认配置文件名称

DEFAULT_CLI_CONFIG_FILE = os.path.join(
    DEFAULT_CONFIG_DIR, __DEFAULT_CLI_CONFIG_FILE_NAME
)  # 默认配置文件路径

__SQLITE_DB_NAME = "dsdl_cli.db"  # sqlite数据库文件

SQLITE_DB_PATH = os.path.join(DEFAULT_CONFIG_DIR, __SQLITE_DB_NAME)  # sqlite数据库文件路径

PROG_NAME = "odl-cli"  # 程序名称

DEFAULT_LOCAL_STORAGE_PATH = os.path.join(DEFAULT_CONFIG_DIR, "datasets")  # 默认本地存储路径

DEFAULT_CLI_LOG_FILE_PATH = os.path.join(DEFAULT_CONFIG_DIR, 'logs') # default log file path


"""
cd function
Examples:
    $python cli.py cd --datasetname coco
    >> now in the Coco dataset context, you can use the following commands:
    >> inspect select studio search
"""

import os
import psutil
import sys
import platform
from distutils.spawn import find_executable
from typing import Dict


from commands.cmdbase import CmdBase
from commons.argument_parser import EnvDefaultVar

from utils.admin import DBClient
import commons.stdio as stdio


class Cd(CmdBase):
    """
    cd command
    """

    def init_parser(self, subparsers):
        """
        Initialize the parser for the command
        document : https://docs.python.org/zh-cn/3/library/argparse.html#
        Args:
            subparsers:
        Returns:
        """
        cd_parser = subparsers.add_parser(
            "cd",
            help="change the context to the specified dataset.",
            example="cd.example",
        )
        cd_parser.add_argument(
            "dataset_name",
            action=EnvDefaultVar,
            envvar="DSDL_CLI_DATASET_NAME",
            type=str,
            help="dataset name",
            metavar="[DATASET NAME]",
        )
        return cd_parser

    def cmd_entry(self, cmdargs, config, *args, **kwargs):
        """
        Entry point for the command
        Args:
            self:
            cmdargs:
            config:
        Returns:
        """

        # _act = _Activator()
        # _act.environ.setdefault("DATASET_NAME", "")
        # _act.environ["DATASET_NAME"] = cmdargs.dataset_name[0]
        # dsname = _act.environ.get("DATASET_NAME", "default")

        dsname = ""

        if cmdargs.dataset_name:
            os.environ.setdefault("DATASET_NAME", "")
            os.environ["DATASET_NAME"] = cmdargs.dataset_name
            dsname = os.environ.get("DATASET_NAME", "default")
        else:
            pass

        dbcli = DBClient()
        is_exists = dbcli.is_dataset_local_exist(dsname)

        if "DATASET_NAME" in os.environ and is_exists:
            sysstr = platform.system()

            if sysstr == "Windows":
                stdio.print_stdout("Enter new Windows cmd command shell")
                shell_cmd = CmdExeActivator().activate_cmd
                os.system(shell_cmd)
            elif sysstr in ["Linux"]:
                stdio.print_stdout("Enter new Linux bash command shell")
                shell_cmd = PosixActivator().activate_cmd
                os.system(shell_cmd)
            elif sysstr in ["Darwin"]:
                stdio.print_stdout("Enter new Darwin bash command shell")
                shell_cmd = PosixActivator().activate_cmd
                os.system(shell_cmd)
            else:
                stdio.print_stderr("Other Systems hava not been supported yet!")
        else:
            stdio.print_stderr("Dataset is not exist, please check the dataset name.")


class _Activator:
    # cd command Activate have three tasks
    #   1. Set and unset environment variables
    #   2. Execute base bash/csh/cmd/powershell base scripts
    #   3. Update the default varible values in the new shell environment

    pathsep_join = None
    sep = None
    script_extension = None
    tempfile_extension = (
        None  # None means write instructions to stdout rather than a temp file
    )
    command_join: str

    run_script_tmpl = None

    def __init__(self, arguments=None):
        self._raw_arguments = arguments
        self.environ = os.environ.copy()
        self.on_linux = bool(sys.platform == "linux")
        self.on_win = bool(sys.platform == "win32")
        self.on_mac = bool(sys.platform == "darwin")
        self.activate_cmd = self._get_activate_cmd()

    def activate(self):
        builder_result = self.build_activate()
        return self._finalize(self._yield_commands(builder_result))

    def _finalize(self, commands):
        merged = {}
        for _cmds in commands:
            merged.update(_cmds)
        commands = merged

    def _yield_commands(self, cmds_dict):
        for key in cmds_dict.get("unset_vars", ()):
            yield self.unset_var_tmpl % key

        for key, value in cmds_dict.get("set_vars", {}).items():
            yield self.set_var_tmpl % (key, value)

        for key, value in cmds_dict.get("export_vars", {}).items():
            yield self.export_var_tmpl % (key, value)

        for script in cmds_dict.get("activate_scripts", ()):
            yield self.run_script_tmpl % script

    def build_activate(self):
        return self._build_activate_stack()

    def _build_activate_stack(self):
        # get environment prefix
        paths = self._get_env_paths()
        activate_cmd = self._get_activate_cmd(paths)

        return {
            # TODO
            # "unset_vars": unset_vars,
            # "set_vars": set_vars,
            # "export_vars": export_vars,
            "activate_scripts": activate_cmd,
        }

    def _get_env_paths(self):
        # include macos and windows
        clean_paths = {
            "darwin": "/usr/bin:/bin:/usr/sbin:/sbin",
            "win32": "C:\\Windows\\system32;"
            "C:\\Windows;"
            "C:\\Windows\\System32\\Wbem;"
            "C:\\Windows\\System32\\WindowsPowerShell\\v1.0\\",
        }
        path = self.environ.get(
            "PATH",
            clean_paths[sys.platform]
            if sys.platform in clean_paths
            else "/usr/bin",  # default is /usr/bin of unix/linux os
        )
        path_split = path.split(os.pathsep)
        return path_split

    def _get_activate_cmd(self) -> str:
        raise NotImplementedError


class PosixActivator(_Activator):
    def __init__(self, arguments=None):
        self.pathsep_join = ":".join
        self.sep = "/"
        self.script_extension = ".sh"
        self.command_join = "\n"

        self.unset_var_tmpl = "unset %s"
        self.set_var_tmpl = "%s='%s'"
        self.export_var_tmpl = "export %s='%s'"
        self.run_script_tmpl = '. "%s"'

        super().__init__(arguments)

    def _get_activate_cmd(self) -> str:
        parent_pid = os.getppid()
        proc = psutil.Process(parent_pid)
        shell_type = proc.name().split(".")[0]
        cmd_location = find_executable(shell_type)
        return f'"{cmd_location}"'


class CmdExeActivator(_Activator):
    def __init__(self, arguments=None):
        self.pathsep_join = ";".join
        self.sep = "\\"
        self.script_extension = ".bat"
        self.tempfile_extension = ".bat"
        self.command_join = "\n"

        self.unset_var_tmpl = "@SET %s="
        self.set_var_tmpl = (
            '@SET "%s=%s"'  # TODO: determine if different than export_var_tmpl
        )
        self.export_var_tmpl = '@SET "%s=%s"'
        self.run_script_tmpl = '@CALL "%s"'

        super().__init__(arguments)

    def _get_activate_cmd(self) -> str:
        parent_pid = os.getppid()
        proc = psutil.Process(parent_pid)
        parent_proc = proc.parent()
        shell_type = parent_proc.name().split(".")[0]
        cmd_location = find_executable(shell_type)
        return f'"{cmd_location}"'


class PowerShellActivator(_Activator):
    def __init__(self, arguments=None):
        self.pathsep_join = ";".join if self.on_win else ":".join
        self.sep = "\\" if self.on_win else "/"
        self.script_extension = ".ps1"
        self.tempfile_extension = (
            None  # write instructions to stdout rather than a temp file
        )
        self.command_join = "\n"

        self.unset_var_tmpl = '$Env:%s = ""'
        self.set_var_tmpl = '$Env:%s = "%s"'
        self.export_var_tmpl = '$Env:%s = "%s"'
        self.run_script_tmpl = '. "%s"'

        super().__init__(arguments)


class CshActivator(_Activator):
    def __init__(self, arguments=None):
        self.pathsep_join = ":".join
        self.sep = "/"
        self.script_extension = ".csh"
        self.tempfile_extension = (
            None  # write instructions to stdout rather than a temp file
        )
        self.command_join = ";\n"

        self.unset_var_tmpl = "unsetenv %s"
        self.set_var_tmpl = "set %s='%s'"
        self.export_var_tmpl = 'setenv %s "%s"'
        self.run_script_tmpl = 'source "%s"'

        super().__init__(arguments)


activator_map: Dict[str, _Activator] = {
    "posix": PosixActivator,
    "ash": PosixActivator,
    "bash": PosixActivator,
    "dash": PosixActivator,
    "zsh": PosixActivator,
    "csh": CshActivator,
    "tcsh": CshActivator,
    "cmd.exe": CmdExeActivator,
    "powershell": PowerShellActivator,
}



import os
import sys
from pathlib import Path
import signal

import commons.stdio as stdio
from commons.exceptions import CLIException, ExistCode
from loguru import logger


class View:
    def __init__(self, dataset_name: str, inspect: bool = False):
        if inspect is True:
            self.dataset_name = dataset_name
        else:
            self.dataset_name = dataset_name.lower()
        self.view_base_dir = Path(os.path.dirname(__file__))
        self.view_code_name = ".".join([self.dataset_name, "py"])
        self.view_code_abspath = Path.joinpath(
            self.view_base_dir, self.dataset_name, self.view_code_name
        )

    def view_from_inspect(self):
        """
        view the files which get from inspect command on webpage.
        Returns: dataset visualization webpage url.
        """
        stdio.print_stdout(
            f"Processing local dataset {self.dataset_name} inspect visulization..."
        )

        view_base_dir = Path(os.path.dirname(__file__))
        view_code_name = "view_from_inspect.py"
        view_code_abspath = Path.joinpath(view_base_dir, view_code_name)
        streamlit_cmd = (
            f"streamlit run {view_code_abspath} -- --dataset-name {self.dataset_name}"
        )
        try:
            signal.signal(signal.SIGINT, self.ctrl_c_handler)
            os.system(streamlit_cmd)
        except CLIException as e:
            stdio.print_stderr(e.message)
            logger.exception(e.message)
            raise CLIException(ExistCode.VIEW_FROM_INSPECT_FAILED, str(e))

    def view_local_dataset(self):
        """
        view the local dataset on webpage.
        Returns: dataset visualization webpage url.
        """
        stdio.print_stdout(
            f"Processing local dataset {self.dataset_name} visulization..."
        )

        for dir_name in self.view_base_dir.iterdir():
            if os.path.isdir(dir_name) and dir_name.name == self.dataset_name:
                for file in dir_name.iterdir():
                    if os.path.exists(self.view_code_abspath):
                        stdio.print_stdout(
                            f"Code for Dataset:{self.dataset_name} visualization has been found."
                        )
                        streamlit_cmd = f"streamlit run {self.view_code_abspath}"
                        try:
                            signal.signal(signal.SIGINT, self.ctrl_c_handler)
                            os.system(streamlit_cmd)
                        except CLIException as e:
                            logger.exception(e.message)
                            stdio.print_stderr(e.message)
                            raise CLIException(
                                ExistCode.VIEW_LOCAL_DATASET_FAILED, str(e)
                            )
                        break
                    else:
                        raise CLIException(
                            ExistCode.VIEW_LOCAL_DATASET_FAILED,
                            "Visulization of Dataset:{self.dataset_name} is to be supported in the near future.",
                        )

    def view_remote_dataset(self):
        """
        view the remote dataset on webpage.
        Returns: dataset visualization webpage url.
        """
        stdio.print_stderr(
            f"[ TO BE DONE ] Processing remote dataset {self.dataset_name} visulization.\n Bye..."
        )

    def ctrl_c_handler(self, signum, frame):
        response = input("Are you sure you want to exit? (y/n): ")
        if response == "y":
            stdio.print_stderr("View cancelled")
            logger.exception("View cancelled")
            sys.exit(0)



"""
admin module handles sqlite operations
"""
import logging
import os
import sqlite3
import json

import pandas as pd
from tabulate import tabulate

from commands import const
from commons.stdio import print_stdout
from loguru import logger
from commons.exceptions import CLIException, ExistCode

# get const path
DB_PATH = const.SQLITE_DB_PATH
DB_DIR_PATH = const.DEFAULT_CONFIG_DIR
DB_NAME = const.__SQLITE_DB_NAME
CONF_PATH = const.DEFAULT_CLI_CONFIG_FILE


def get_config_dict():
    with open(CONF_PATH, 'r') as f:
        return json.loads(f.read())


def __get_default_path() -> str:
    """
    Get the default DSDL directory path
    @return: default DSDL directory path as string
    """
    path = DB_DIR_PATH
    if not os.path.exists(path):
        print("initialize default directory...")
        os.mkdir(path)
    return path


def __get_default_db_path():
    """
    Get the default DSDL sqlite file path
    @return:
    """
    path = os.path.join(__get_default_path(), DB_NAME)
    if not os.path.exists(path):
        print("initialize default db file...")
        initialize_db(path)
    return path


def initialize_db(db_file):
    """
    Initialize sqlite db file in defautl DSDL folder
    @param db_file: default db file path
    @return:
    """
    conn = sqlite3.connect(database=db_file)
    cursor = conn.cursor()
    create_table_sql = '''
    CREATE TABLE IF NOT EXISTS dataset(
    dataset_name varchar, 
    storage_name varchar,
    dataset_path varchar,
    label_data boolean,
    media_data boolean,
    dataset_media_file_num bigint,
    dataset_media_file_bytes bigint,
    created_time timestamp,
    updated_time timestamp,
    primary key(dataset_name)
    );
    '''
    cursor.execute(create_table_sql)

    create_table_sql = '''
    CREATE TABLE IF NOT EXISTS split(
    dataset_name varchar, 
    split_name varchar,
    split_type varchar,
    label_data boolean,
    media_data boolean,
    split_media_file_num bigint,
    split_media_file_bytes bigint, 
    created_time timestamp,
    updated_time timestamp,
    primary key(dataset_name, split_name),
    CONSTRAINT fk_dataset  
    FOREIGN KEY (dataset_name)  
    REFERENCES departments(dataset_name)  
    );
    '''
    cursor.execute(create_table_sql)

    conn.commit()
    cursor.close()
    conn.close()


def get_size_sum(file_list):
    """
    Compute total size of a list of files
    @param file_list: a list of file path
    @return: the total size of files
    """
    sum_size = 0
    for f in file_list:
        sum_size += os.path.getsize(f)
    return sum_size


def get_media_stat(media_list):
    """
    Computer statistics of a list of media files
    @param media_list:
    @return: media statistics
    """
    media_list = set(media_list)
    media_num = len(media_list)
    media_size = get_size_sum(media_list)

    media_stat = {'media_num': media_num, 'media_size': media_size}
    return media_stat


class DBClient:
    """
    This class handles operations on DSDL local db
    """

    def __init__(self):
        """
        create connection and cursor to link sqlite
        """
        self.conn = sqlite3.connect(database=DB_PATH)
        self.cursor = self.conn.cursor()

    # def __del__(self):
    #     """
    #     close connection and cursor
    #     @return:
    #     """
    #     self.cursor.close()
    #     self.conn.close()
    def get_sqlite_dict_list(self, sql) -> list:
        """
        Get sql query result as a list of dicts
        @param cursor: a cursor of a db connection
        @param sql: query sql
        @return: the query result as a list of dicts
        """
        cursor = self.cursor
        res_list = []
        res = cursor.execute(sql).fetchall()
        header = [x[0] for x in cursor.description]
        for r in res:
            res_list.append(dict(zip(header, r)))
        return res_list

    def get_sqlite_dataframe(self, sql) -> pd.DataFrame:
        """
        Get sql query result as a pandas dataframe
        @param cursor: a cursor of a db connection
        @param sql: query sql
        @return: the query result as a pandas dataframe
        """
        dict_list = self.get_sqlite_dict_list(sql)
        dataframe = pd.DataFrame.from_dict(data=dict_list)
        return dataframe

    def get_local_dataset_path(self, dataset_name: str):
        """
        Get the local path from sqlite for the given dataset name

        @param dataset_name: the formal dataset name which you want to get local storage path
        @return: the dataset local path get from sqlite db
                 return None if there is no record in database for the given dataset name
        """
        try:
            res = self.cursor.execute("select dataset_path from dataset where dataset_name=?",
                                      [dataset_name]).fetchone()
        except Exception as e:
            error_info = "Sqlite select dataset operation error: \n" + str(e)
            logger.error(error_info)
            raise CLIException(ExistCode.SQLITE_OPERATION_ERROR, error_info)
        if res:
            return res[0]
        else:
            return None

    def get_local_split_path(self, dataset_name, split_name):
        """
        Get the local path from sqlite for the given dataset name and split name
        @param dataset_name: the formal dataset name
        @param split_name: the split of the dataset, such as train/test
        @return: the split local path get from sqlite db
                 return None if there is no record in database for the given dataset name
        """
        dataset_path = self.get_local_dataset_path(dataset_name)
        try:
            split_data = self.cursor.execute("select * from split where dataset_name=? and split_name=?",
                                             [dataset_name, split_name]).fetchone()
        except Exception as e:
            error_info = "Sqlite select split operation error: \n" + str(e)
            logger.error(error_info)
            raise CLIException(ExistCode.SQLITE_OPERATION_ERROR, error_info)
        split_path = os.path.join(dataset_path, 'parquet', '%s.parquet' % split_name)
        if split_data and os.path.exists(split_path):
            return split_path
        else:
            return None

    def is_dataset_local_exist(self, dataset_name: str) -> bool:
        """
        Check whether the given dataset exists locally

        @param dataset_name: the formal dataset name which you want to check if exists locally
        @return: if exists, return True, otherwise return False
        """
        if self.get_local_dataset_path(dataset_name):
            return True
        else:
            return False

    def is_split_local_exist(self, dataset_name: str, split_name: str) -> bool:
        """
        Check whether the given split exists locally
        @param dataset_name: the formal dataset name which the target split belong to
        @param split_name: the split name which you want to check if exists locally
        @return: if exists, return True, otherwise return False
        """
        try:
            res = self.cursor.execute("select * from split where dataset_name=? and split_name=?",
                                      [dataset_name, split_name]).fetchone()
        except Exception as e:
            error_info = "Sqlite select split operation error: \n" + str(e)
            logger.error(error_info)
            raise CLIException(ExistCode.SQLITE_OPERATION_ERROR, error_info)
        if res:
            return True
        else:
            return False

    def register_dataset(self, dataset_name, storage_name, dataset_path, label, media, media_num, media_size):
        """
        Register a dataset in database
        @param dataset_name: the dataset name
        @param storage_name: the storage name in conf file
        @param dataset_path: the dataset storage path
        @param label: 1 or 0, whether the label data of dataset is downloaded
        @param media: 1 or 0, whether the media data of dataset is downloaded
        @param media_num: the number of media files
        @param media_size: the number of total media file size
        @return:
        """
        try:
            self.cursor.execute(
                "insert or replace into dataset values (?,?,?,?,?,?,?,datetime('now','localtime'),datetime('now','localtime'))",
                [dataset_name, storage_name, dataset_path, label, media, media_num, media_size])
            self.conn.commit()
        except Exception as e:
            error_info = "Sqlite register dataset operation error: \n" + str(e)
            logger.error(error_info)
            raise CLIException(ExistCode.SQLITE_OPERATION_ERROR, error_info)

    def register_split(self, dataset_name, split_name, split_type, label, media, media_num, media_size):
        """
        Register a new split in database
        @param dataset_name: the dataset name
        @param split_name: a subset of a dataset
        @param media_num: the number of media files
        @param media_size: the number of total media file size
        @return:
        """
        try:
            self.cursor.execute(
                "insert or replace into split values (?,?,?,?,?,?,?,datetime('now','localtime'),datetime('now','localtime'))",
                [dataset_name, split_name, split_type, label, media, media_num, media_size])
            self.conn.commit()
        except Exception as e:
            error_info = "Sqlite register split operation error: \n" + str(e)
            logger.error(error_info)
            raise CLIException(ExistCode.SQLITE_OPERATION_ERROR, error_info)

    def delete_split(self, dataset_name, split_name):
        """
        Delete a split from database after it was deleted
        @param dataset_name: the dataset name
        @param split_name: a subset of a dataset
        @return:
        """
        try:
            self.cursor.execute(
                "delete from split where dataset_name=? and split_name=?",
                [dataset_name, split_name])
            self.conn.commit()
        except Exception as e:
            error_info = "Sqlite delete split operation error: \n" + str(e)
            logger.error(error_info)
            raise CLIException(ExistCode.SQLITE_OPERATION_ERROR, error_info)

    def delete_dataset(self, dataset_name):
        """
        Delete a dataset from database after it was deleted
        @param dataset_name: the dataset name
        @return:
        """
        try:
            self.cursor.execute(
                "delete from split where dataset_name=?",
                [dataset_name])
            self.cursor.execute(
                "delete from dataset where dataset_name=?",
                [dataset_name])
            self.conn.commit()
        except Exception as e:
            error_info = "Sqlite delete dataset operation error: \n" + str(e)
            logger.error(error_info)
            raise CLIException(ExistCode.SQLITE_OPERATION_ERROR, error_info)

    def is_dataset_label_downloaded(self, dataset_name):
        """
        Whether the label data of a dataset is downloaded
        @param dataset_name: dataset name
        @return:
        """
        try:
            res = self.cursor.execute("select label_data from dataset where dataset_name=?",
                                      [dataset_name]).fetchone()
        except Exception as e:
            error_info = "Sqlite select dataset operation error: \n" + str(e)
            logger.error(error_info)
            raise CLIException(ExistCode.SQLITE_OPERATION_ERROR, error_info)
        flag = False
        if res:
            if res[0] == 1:
                flag = True

        return flag

    def is_dataset_media_downloaded(self, dataset_name):
        """
        Whether the media data of a dataset is downloaded
        @param dataset_name:
        @return:
        """
        try:
            res = self.cursor.execute("select media_data from dataset where dataset_name=?",
                                      [dataset_name]).fetchone()
        except Exception as e:
            error_info = "Sqlite select dataset operation error: \n" + str(e)
            logger.error(error_info)
            raise CLIException(ExistCode.SQLITE_OPERATION_ERROR, error_info)
        flag = False
        if res:
            if res[0] == 1:
                flag = True

        return flag


if __name__ == '__main__':
    print(DB_PATH)
    print(DB_DIR_PATH)
    default_path = __get_default_path()
    print(default_path)
    db_path = __get_default_db_path()
    print(db_path)
    # print(get_local_dataset_path('CIFAR-10'))
    # print(get_local_split_path('CIFAR-100', 'test'))
    db_client = DBClient()
    print(db_client.get_sqlite_dict_list('select * from dataset'))
    # print(get_sqlite_dataframe('select * from dataset', get_sqlite_table_header('dataset')))
    df = db_client.get_sqlite_dataframe('select * from dataset')
    print(tabulate(df, headers='keys', tablefmt='psql', showindex=False))
    # print(db_client.is_split_local_exist('CIFAR-10', 'test2'))
    print(db_client.is_dataset_label_downloaded('CIFAR-10'))
    print(db_client.is_dataset_media_downloaded('CIFAR-10'))
    print(get_config_dict())



""""
定义输出到控制台的函数。
包含标准流输出、错误流输出
"""
import sys


def print_stdout(msg, *args, sep=' ', end='\n'):
    """标准输出"""
    print(msg, *args, sep=sep, end=end, file=sys.stdout)


def print_stderr(msg, *args, sep=' ', end='\n'):
    """错误输出"""
    print(msg, *args, sep=sep, end=end, file=sys.stderr)



from typing import List

import cv2
import numpy as np

from src.types import Array, Color

WHITE = (255, 255, 255)


def text_in_rectangle(  # pylint: disable=too-many-arguments
    image: Array,
    text: str,
    x1: int,
    y1: int,
    color: Color,
    font_scale=0.75,
    thickness=2,
    font_face=cv2.FONT_HERSHEY_DUPLEX,
    text_color: Color = WHITE,
) -> Array:
    (text_w, text_h), _ = cv2.getTextSize(
        text, fontFace=font_face, fontScale=font_scale, thickness=thickness
    )
    image = cv2.rectangle(
        image, (x1, y1), (x1 + text_w + 2, y1 - text_h - 4), color, -1
    )
    image = cv2.putText(
        image,
        text,
        (x1 + 1, y1 - 2),
        fontFace=font_face,
        fontScale=font_scale,
        color=text_color,
        thickness=thickness,
    )
    return image


def draw_bboxes(
    image: Array,
    bboxes: Array,
    colors: List[Color],
    class_names: List[str],
    alpha=0.2,
) -> Array:
    processed_image = image.copy()
    for bbox, color in zip(bboxes, colors):
        height, width = bbox[3] - bbox[1], bbox[2] - bbox[0]
        canvas = image.copy()
        patch = np.full((height, width, 3), color, dtype=np.uint8)
        canvas[bbox[1] : bbox[3], bbox[0] : bbox[2]] = patch
        processed_image = cv2.addWeighted(
            processed_image, 1 - alpha, canvas, alpha, 0.0
        )
    for bbox, color in zip(bboxes, colors):
        processed_image = cv2.rectangle(
            processed_image, (bbox[0], bbox[1]), (bbox[2], bbox[3]), color, 2
        )
    font_scale = min(image.shape[:2]) / 1200 + 0.5
    thickness = round(font_scale * 2 + 0.5)
    for bbox, color, class_name in zip(bboxes, colors, class_names):
        processed_image = text_in_rectangle(
            processed_image,
            class_name,
            bbox[0],
            bbox[1],
            font_scale=font_scale,
            thickness=thickness,
            color=color,
        )
    return processed_image



# import the Package
import os
import sys
import base64
from io import BytesIO
import argparse
import glob

sys.path.append(f"{os.path.dirname(os.path.abspath(__file__))}/../../")

import streamlit as st
from PIL import Image
import pandas as pd

from utils.query import SplitReader
from utils.admin import DBClient


def main():
    parser = argparse.ArgumentParser(description="streamlit app inside odl-cli.")

    parser.add_argument(
        "--dataset-name",
    )
    parser.add_argument(
        "--split-name",
        default="train",
    )

    try:
        args = parser.parse_args()
    except SystemExit as e:
        os._exit(e.code)

    dataset_name = args.dataset_name
    split_name = args.split_name
    st.title("Files")
    # explore_app(dataset_name, split_name)
    image_grid_main(dataset_name, split_name)


def explore_app(dataset_name: str, split_name: str):
    dataset_name = dataset_name
    split_name = "train"  # currently only support train split
    print(f"dataset_name: {dataset_name}, split_name: {split_name}")

    db_client = DBClient()
    path = db_client.get_local_split_path(dataset_name, split_name)
    print(path)

    split_reader = SplitReader(dataset_name, split_name)
    print(split_reader.dataset_name, split_reader.split_name)
    SPLIT_DF = pd.DataFrame()

    SPLIT_DF["file_path"] = split_reader.get_image_samples(10)
    SPLIT_DF["split"] = split_name
    SPLIT_DF["thumbnail"] = SPLIT_DF.file_path.map(lambda x: image_formatter(x))
    DF = pd.DataFrame(
        {
            "thumbnail": SPLIT_DF["thumbnail"],
            "split": SPLIT_DF["split"],
        }
    )

    DF_HTML = convert_df(DF)

    pd.set_option("display.max_colwidth", -1)
    st.markdown(DF_HTML, unsafe_allow_html=True)


def image_grid_main(dataset_name: str, split_name: str):
    st.title("Image Grid Display")
    image_files, manuscripts = load_images(dataset_name, split_name)
    view_manuscripts = st.multiselect("Select Manuscript(s)", manuscripts)
    n = st.number_input("Select Grid Width", 1, 5, 3)

    view_images = []
    for image_file in image_files:
        if any(manuscript in image_file for manuscript in view_manuscripts):
            view_images.append(image_file)
    groups = []
    for i in range(0, len(view_images), n):
        groups.append(view_images[i : i + n])

    for group in groups:
        cols = st.columns(n)
        for i, image_file in enumerate(group):
            cols[i].image(image_file)


# @st.cache
def load_images(dataset_name: str, split_name: str):
    files = get_file_list(dataset_name, split_name)
    image_files = files
    manuscripts = []
    for image_file in image_files:
        image_file = image_file.replace("\\", "/")
        parts = image_file.split("/")
        if parts[1] not in manuscripts:
            manuscripts.append(parts[1])
    manuscripts.sort()

    return image_files, manuscripts


def get_file_list(dataset_name: str, split_name: str, count=10):
    dataset_name = dataset_name
    split_name = "train"  # currently only support train split
    print(f"dataset_name: {dataset_name}, split_name: {split_name}")

    split_reader = SplitReader(dataset_name, split_name)
    print(split_reader.dataset_name, split_reader.split_name)

    files = split_reader.get_image_samples(count)
    return files


@st.cache
def convert_df(input_df):
    # IMPORTANT: Cache the conversion to prevent computation on every rerun
    return input_df.to_html(
        formatters={"img": image_formatter},
        escape=False,
    )


def get_thumbnail(path):
    i = Image.open(path)
    i.thumbnail((1280, 1280), Image.LANCZOS)
    return i


def image_base64(im):
    if isinstance(im, str):
        im = get_thumbnail(im)
    with BytesIO() as buffer:
        im.save(buffer, "jpeg")
        return base64.b64encode(buffer.getvalue()).decode()


def image_formatter(im):
    return f'<img src="data:image/jpeg;base64,{image_base64(im)}">'


if __name__ == "__main__":
    main()



"""
oss ops module handles s3 operations
"""
import logging
import os
import time

from boto3.session import Session
from botocore.exceptions import ClientError

from commons.stdio import print_stdout
from loguru import logger
from commons.exceptions import CLIException, ExistCode


def print_progress(iteration, total, start_time, prefix='', suffix='', decimals=1, bar_length=100):
    """
    Call in a loop to create a terminal progress bar
    @param iteration:   - Required  : current iteration (Int)
    @param total:       - Required  : total iterations (Int)
    @param start_time:  - Required  : process start time (Float)
    @param prefix:      - Optional  : prefix string (Str)
    @param suffix:      - Optional  : suffix string (Str)
    @param decimals:    - Optional  : positive number of decimals in percent complete (Int)
    @param bar_length:  - Optional  : character length of bar (Int)
    @return:
    """
    import sys
    format_str = "{0:." + str(decimals) + "f}"
    percent = format_str.format(100 * (iteration / float(total)))
    filled_length = int(round(bar_length * iteration / float(total)))
    bar = '#' * filled_length + '-' * (bar_length - filled_length)
    dur = time.perf_counter() - start_time
    sys.stdout.write('\r%s |%s| %ds %s%s %s' % (prefix, bar, dur, percent, '%', suffix)),
    if iteration == total:
        sys.stdout.write('\n')
    sys.stdout.flush()


class OssClient:
    """
    S3 connection client
    """

    def __init__(self, aws_access_key_id, aws_secret_access_key, endpoint_url, region_name):
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.endpoint_url = endpoint_url
        self.region_name = region_name

        self.session = Session(aws_access_key_id, aws_secret_access_key)
        self.s3_client = self.session.client("s3", endpoint_url=endpoint_url, region_name=region_name, use_ssl=False)

    def list_buckets(self):
        """
        List all the bucket names
        @return: A list of bucket names
        """
        response = dict(self.s3_client.list_buckets())
        bucket_list = [x['Name'] for x in response['Buckets']]
        return bucket_list

    def list_objects(self, bucket, prefix, delimiter=''):
        """
        List all objects in a given bucket start with a given prefix
        @param bucket:
        @param prefix:
        @param delimiter:
        @return:
        """
        obj_list = []
        paginator = self.s3_client.get_paginator('list_objects_v2')
        pages = paginator.paginate(Bucket=bucket, Prefix=prefix, Delimiter=delimiter)

        for page in pages:
            if 'Contents' in page.keys():
                for obj in page['Contents']:
                    obj_list.append(obj)
        return obj_list

    def get_dir_list(self, bucket, remote_directory):
        """
        List directories in a remote directory
        @param bucket:
        @param remote_directory:
        @return:
        """
        child_dir_list = set()
        paginator = self.s3_client.get_paginator('list_objects_v2')
        pages = paginator.paginate(Bucket=bucket, Prefix=remote_directory, Delimiter='/')

        for page in pages:
            if 'CommonPrefixes' in page.keys():
                for p in page['CommonPrefixes']:
                    child_dir_list.add(p['Prefix'])

        return sorted(child_dir_list)

    def __get_recursive_dir(self, dirs, bucket, remote_directory):
        """
        A recursive func to get all directories in a remote directory
        @param dirs:
        @param bucket:
        @param remote_directory:
        @return:
        """
        dirs.append(remote_directory)
        # child_list = set()
        # paginator = self.s3_client.get_paginator('list_objects_v2')
        # pages = paginator.paginate(Bucket=bucket, Prefix=remote_directory, Delimiter='/')
        #
        # for page in pages:
        #     if 'CommonPrefixes' in page.keys():
        #         for p in page['CommonPrefixes']:
        #             child_list.add(p['Prefix'])

        child_dir_list = self.get_dir_list(bucket, remote_directory)
        if len(child_dir_list) != 0:
            for child_dir in child_dir_list:
                self.__get_recursive_dir(dirs, bucket, child_dir)

    def get_recursive_dir_list(self, bucket, remote_directory):
        """
        Call a recursive func to get all directories in a remote directory
        @param bucket:
        @param remote_directory:
        @return:
        """
        dir_list = []
        self.__get_recursive_dir(dir_list, bucket, remote_directory)
        dir_list = [dir[len(remote_directory):] for dir in dir_list]
        return dir_list

    def download_file(self, bucket, remote_file, local_file):
        """
        download a file from s3
        :param bucket: bucket name
        :param remote_file: remote file name
        :param local_file: local file name
        :return:
        """
        try:
            if not os.path.exists(local_file):
                self.s3_client.download_file(bucket, remote_file, local_file)
        except ClientError as e:
            logging.error(e)
            return False
        return True

    def download_directory(self, bucket, remote_directory, local_directory):
        """
        download a directory from s3
        :param bucket: bucket name
        :param remote_directory: remote directory name
        :param local_directory: local directory name
        :return:
        """
        print_stdout("preparing...")
        dirs_list = self.get_recursive_dir_list(bucket=bucket, remote_directory=remote_directory)
        for d in dirs_list:
            path = os.path.join(local_directory, d)
            if not os.path.exists(path):
                os.mkdir(path)

        print_stdout("start download...")

        # print(bucket)
        # print(remote_directory)
        file_list = self.list_objects(bucket, remote_directory)
        process = 0
        file_number = len(file_list)
        start_time = time.perf_counter()

        print_progress(process, file_number, start_time, prefix='Downloading', suffix='Complete',
                       bar_length=50)

        for file in file_list:
            file_name = file['Key']
            local_file = os.path.join(local_directory, file_name[len(remote_directory):])
            if not os.path.exists(local_file):
                self.download_file(bucket, file_name, local_file)
            process += 1
            print_progress(process, file_number, start_time, prefix='Download', suffix='Complete',
                           bar_length=50)

        print_stdout('Download Complete')

    def download_list(self, bucket, media_list, remote_directory, local_directory):
        """
        Download a directory from s3
        @param bucket: bucket name
        @param media_list: a list of media files to download
        @param remote_directory: remote directory name
        @param local_directory: local directory name
        @return:
        """
        print_stdout("preparing...")
        dirs_list = self.get_recursive_dir_list(bucket=bucket, remote_directory=remote_directory)
        # print(dirs_list)
        for d in dirs_list:
            path = os.path.join(local_directory, d)
            if not os.path.exists(path):
                os.mkdir(path)

        print_stdout("start download...")

        process = 0
        file_number = len(media_list)
        start_time = time.perf_counter()

        print_progress(process, file_number, start_time, prefix='Downloading', suffix='Complete',
                       bar_length=50)

        for media in media_list:
            file_key = remote_directory + media
            local_file = os.path.join(local_directory, media)
            if not os.path.exists(local_file):
                self.download_file(bucket, file_key, local_file)
            process += 1
            print_progress(process, file_number, start_time, prefix='Download', suffix='Complete',
                           bar_length=50)

        print_stdout('Download Complete')

    def get_sum_size(self, bucket, file_key_list):
        sum = 0
        for key in file_key_list:
            data = self.s3_client.head_object(Bucket=bucket, Key=key)
            size = int(data['ResponseMetadata']['HTTPHeaders']['content-length'])
            sum += size
        return sum

    def read_file(self, bucket, remote_file):
        """

        @param bucket:
        @param remote_file:
        @return:
        """
        data = self.s3_client.get_object(Bucket=bucket, Key=remote_file)
        contents = data['Body'].read()
        return contents.decode("utf-8")

    def list_datasets(self, bucket):
        """
        Return a list of dataset names
        @return:
        """
        dataset_list = [x.replace('/', '') for x in self.get_dir_list(bucket, '')]
        return dataset_list

    def is_dataset_remote_exist(self, bucket, dataset_name):
        """

        @param bucket:
        @param dataset_name:
        @return:
        """
        dataset_list = self.list_datasets(bucket)
        if dataset_name in dataset_list:
            return True
        else:
            return False

    def list_splits(self, bucket, dataset_name):
        """

        @param bucket:
        @param dataset_name:
        @return:
        """
        if not self.is_dataset_remote_exist(bucket, dataset_name):
            print_stdout("The dataset %s does not exist in remote repo" % dataset_name)
            # logger.info("The dataset %s does not exist in remote repo" % dataset_name)
            exit()
        else:
            parquet_prefix = dataset_name + "/parquet/"
            parquet_name_list = [x['Key'][len(parquet_prefix):].replace('.parquet', '') for x in
                                 self.list_objects(bucket, parquet_prefix) if str(x['Key']).endswith(".parquet")]
            return parquet_name_list

    def is_split_remote_exist(self, bucket, dataset_name, split_name):
        """

        @param bucket:
        @param dataset_name:
        @param split_name:
        @return:
        """
        split_list = self.list_splits(bucket, dataset_name)
        if split_name in split_list:
            return True
        else:
            return False

    def upload_file(self, local_file, bucket, remote_file=None):
        """
        upload a local file to s3
        :param local_file: file to upload
        :param bucket: bucket to upload to
        :param remote_file: destination file name to be saved
        :return:
        """
        if remote_file is None:
            remote_file = local_file
        try:
            self.s3_client.upload_file(local_file, bucket, remote_file, ExtraArgs={"ACL": "public-read"})
        except ClientError as e:
            logging.error(e)
            return False
        return True

    # def list_objects(self, bucket, remote_directory=None):
    #     """
    #     list objects in a bucket
    #     :param bucket:
    #     :return: file list
    #     """
    #     file_list = []
    #     response_generator = self._list_obj_scluster(bucket, remote_directory)
    #     for content in response_generator:
    #         file_list.append(content["Key"])
    #     return file_list

    # def _list_obj_scluster(self, bucket, remote_directory=None):
    #     """
    #     list objects in a bucket
    #     :param bucket:
    #     :return: file_list [file_name]
    #     """
    #     marker = None
    #     file_list = []
    #
    #     while True:
    #         list_kwargs = dict(MaxKeys=1000, Bucket=bucket, Prefix=remote_directory)
    #         if marker:
    #             list_kwargs['Marker'] = marker
    #         response = self.s3_client.list_objects(**list_kwargs)
    #         file_desc = response.get("Contents", [])
    #         yield from file_desc
    #         if not response.get("IsTruncated") or len(file_desc) == 0:
    #             break
    #         marker = file_desc[-1]['Key']

    def obj_is_exist(self, bukcet, obj_name):
        '''
        check if a s3 object already  exists
        :param bukcet: bucket name
        :param obj_name: object name
        :return: True if exists, else False
        '''
        try:
            self.s3_client.head_object(Bucket=bukcet, Key=obj_name)
        except ClientError:
            return False
        return True

    # def download_directory(self, bucket, remote_directory, local_directory):
    #     """
    #     download a directory from s3
    #     :param bucket: bucket name
    #     :param remote_directory: remote directory name
    #     :param local_directory: local directory name
    #     :return:
    #     """
    #
    #     if not os.path.exists(local_directory):
    #         os.makedirs(local_directory)
    #
    #     try:
    #         file_list = self.list_objects(bucket, remote_directory)
    #         for file in file_list:
    #             if file.startswith(remote_directory):
    #                 file_name = file.replace(remote_directory, "").replace("/", "")
    #                 local_file = os.path.join(local_directory, file_name)
    #                 self.download_file(bucket, file, local_file)
    #     except ClientError as e:
    #         logging.error(e)
    #         return False
    #     return True

    def create_directory(self, bucket, directory):
        """
        create a directory in s3
        :param bucket: bucket name
        :param directory: directory name
        :return:
        """
        try:
            self.s3_client.put_object(Bucket=bucket, Key=(directory + "/"))
        except ClientError as e:
            logging.error(e)
            return False
        return True


if __name__ == '__main__':
    aws_access_key_id = "ailabminio"
    aws_secret_access_key = "123123123"
    endpoint_url = "http://10.140.0.94:9800"
    region_name = "ailab"

    s3_client = OssClient(aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key,
                          endpoint_url=endpoint_url, region_name=region_name)

    # print(s3_client.list_buckets())
    # obj_list = s3_client.list_objects('testdata', 'test_data/', '/')
    # print(len(obj_list))
    # print(get_dir_list(obj_list))
    # for obj in obj_list:
    #     print(obj)
    # t = ['test_data/', 'test_data/1/', 'test_data/1/2/', 'test_data/1/2/4/', 'test_data/1/5/', 'test_data/3/',
    #      'test_data/3/6/', 'test_data/3/6/7/']

    # response = s3_client.s3_client.list_objects_v2(Bucket='dsdldata', Prefix='CIFAR-10/', Delimiter='/')
    # print(response)
    # for prefix in response['CommonPrefixes']:
    #     print(prefix['Prefix'][:-1])

    # s3_client.download_directory('dsdldata', 'Fashion-MNIST/', 'D:\\DSDL_DATA')

    # print(s3_client.get_recursive_dir_list('testdata', 'test_data/'))
    # print(len(s3_client.list_objects('dsdldata', 'CIFAR-10/media/')))
    # obj_list = s3_client.read_file('dsdldata', 'CIFAR-10/parquet/dataset.yaml')
    # print(obj_list)
    # print(s3_client.list_datasets("dsdldata"))
    # print(s3_client.list_splits("dsdldata", "STL-10"))
    # print(s3_client.is_split_remote_exist("dsdldata", "STL-101", "unlabeled1"))
    print(s3_client.get_sum_size("dsdldata", ['CIFAR-10-Auto/media/000000007605.png',
                                              'CIFAR-10-Auto/media/000000009822.png']))
    # print(s3_client.get_dir_list('dsdldata', ''))



"""
一个命令的实现例子

Examples:
    $python cli.py rm  -y  openmmlab/flying3d

"""

from commands.cmdbase import CmdBase
from commons.exceptions import CLIException, ExistCode
from commons.stdio import print_stdout
from commons.storage_builder import StorageBuilder
from utils.admin import DBClient

from dsdlsdk.exception.exception import DatasetPathNotExists
from dsdlsdk.storage import Storage


class Rm(CmdBase):
    def __init__(self):
        self.__config = None

    def init_parser(self, subparsers):
        """
        Initialize the parser for the command
        document : https://docs.python.org/zh-cn/3/library/argparse.html#

        Args:
            subparsers:

        Returns:

        """
        rm_parser = subparsers.add_parser(
            "rm", help="Remove dataset in a desired manner", example="rm.example"
        )  # example 样例文件位于resources/下，普通的文本文件，每个命令写一个
        rm_parser.add_argument(
            "-y",
            "--yes",
            default=False,
            help="remove without confirmation",
            action="store_true",
        )
        rm_parser.add_argument(
            "--storage", default="default", type=str, help="local storage name"
        )
        rm_parser.add_argument("dataset_name", type=str, help="dataset name")
        return rm_parser

    def cmd_entry(self, cmdargs, config, *args, **kwargs):
        """
        Entry point for the command

        Args:
            self:
            cmdargs:
            config:

        Returns:

        """
        self.__config = config
        yes = cmdargs.yes
        storage_name = cmdargs.storage
        dataset_name = cmdargs.dataset_name
        try:
            self.__process_dataset_del(dataset_name, storage_name, force_delete=yes)
        except DatasetPathNotExists as e:
            raise CLIException(ExistCode.DATASET_PATH_NOT_EXISTS, str(e))

    def __process_dataset_del(
        self, dataset_name, storage_name, force_delete=False
    ) -> None:
        """
        Delete one dataset

        Args:
            dataset_name:
            storage: local storage name
            force_delete:

        Returns:

        """
        dbcli = DBClient()
        dataset_path = dbcli.get_local_dataset_path(dataset_name)
        # 先查找这个数据集是否存在本地
        if not dataset_path:
            raise CLIException(
                ExistCode.DATASET_NOT_EXIST_LOCAL,
                "Dataset `{}` does not exist locally".format(dataset_name),
            )
        else:  # 存在，那么就删除
            if force_delete:
                self.__delete_dataset(dataset_name, dataset_path, storage_name)
            else:
                print_stdout(
                    "Are you sure to delete dataset {}? [y/n]".format(dataset_name),
                    end="",
                )
                answer = input()
                if answer == "y":
                    self.__delete_dataset(dataset_name, dataset_path, storage_name)
                    dbcli.delete_dataset(dataset_name)
                else:
                    print_stdout("Delete cancelled")

    def __delete_dataset(
        self, dataset_name: str, dataset_path: str, storage_name: str
    ) -> None:
        """
        delete dataset from  storage

        Args:
            dataset_path:
            storage_name:

        Returns:

        """
        dbcli = DBClient()
        storage_cli = self.__get_storage_cli_by_name(storage_name)
        if storage_cli.remove_tree(dataset_path):
            dbcli.delete_dataset(dataset_name)
            print_stdout("Dataset {} deleted".format(dataset_path))

    def __get_storage_cli_by_name(self, storage_name) -> Storage:
        """
        Get storage client by name

        Args:
            storage_name:

        Returns:

        """
        s: Storage = StorageBuilder.build_by_name(storage_name, self.__config)
        return s



from enum import Enum


class ExistCode(Enum):
    """
    定义错误码
    """
    SUCCESS = 0
    NOT_SUPPORTED_STORE_PATH = 1
    STORAGE_NAME_NOT_EXIST = 2  # 有storage这个配置，但是没有storage_name这个配置
    STORAGE_NOT_EXIST = 3  # 没有storage这个配置
    DATASET_NOT_EXIST_LOCAL = 4  # 本地没有这个数据集
    DATASET_PATH_NOT_EXISTS = (
        5  # 虽然数据库里能找到数据集名字，但是路径不存在。比如路径不在默认路径下，需要在删除时候指定 --storage, 否则只会到default路径下
    )
    VIEW_FROM_INSPECT_FAILED = 6  # 从inspect命令中中查看数据集失败
    VIEW_LOCAL_DATASET_FAILED = 7  # 从本地查看数据集失败
    VIEW_REMOTE_DATASET_FAILED = 8  # 从远程查看数据集失败
    QUERY_SYNTAX_ERROR = 9  # 查询数据集时的语法报错
    DATASET_PARAM_NOT_EXIST = 10  # 缺少dataset name参数
    SPLIT_NOT_EXIST = 11  # split在本地和远端都不存在
    DATASET_NOT_EXIST_REMOTE = 12  # dataset在远端都不存在
    SQLITE_OPERATION_ERROR = 13  # sqlite操作报错
    DATASET_NOT_EXIST = 14  # dataset在本地和远端都不存在


class CLIException(Exception):
    """Base exception for all exceptions in this module."""

    def __init__(self, errcode: ExistCode, message: str):
        self.message = message
        self.errcode = int(errcode.value)



"""
admin module handles parquet operations
"""
import os.path
import duckdb
import yaml

from utils import admin
import pyarrow.parquet as pq
import pyarrow as pa
from utils.oss_ops import ops
from pyarrow import fs

aws_access_key_id = "ailabminio"
aws_secret_access_key = "123123123"
endpoint_url = "http://10.140.0.94:9800"
region_name = "ailab"
default_bucket = "dsdldata"


def get_dataset_info(dataset_name):
    db_client = admin.DBClient()
    if not db_client.is_dataset_local_exist(dataset_name):
        print("there is no dataset named %s locally" % dataset_name)
        exit()
    yml_path = os.path.join(db_client.get_local_dataset_path(dataset_name), 'parquet', 'dataset.yaml')
    with open(yml_path, 'r') as f:
        dataset_dict = yaml.safe_load(f)
    return dataset_dict


class ParquetReader:
    """
    A Class to read parquet file
    """

    def __init__(self, parquet_path, endpoint=None, access_key_id=None, secret_access_key=None, url_stype='path',
                 use_ssl=False):
        """
        Construct the parquet reader based on the parquet file path
        @param parquet_path:
        """
        self.cursor = duckdb.connect(database=':memory:')
        if not parquet_path.lower().startswith("s3"):
            self.path_flag = "local"
            self.parquet_path = parquet_path
        else:
            if endpoint is None:
                print("s3 info lacks endpoint")
                exit()
            if access_key_id is None:
                print("s3 info lacks access_key_id")
                exit()
            if secret_access_key is None:
                print("s3 info lacks secret_access_key")
                exit()

            self.endpoint = endpoint
            self.access_key_id = access_key_id
            self.secret_access_key = secret_access_key
            self.url_stype = url_stype
            self.use_ssl = "true" if use_ssl else "false"
            self.path_flag = "s3"
            self.parquet_path = parquet_path
            self.parquet_key = parquet_path[5:]

            self.fs = fs.S3FileSystem(
                access_key=aws_access_key_id,
                secret_key=secret_access_key,
                endpoint_override=endpoint_url,
            )

    def __create_dataset_view(self):
        if self.path_flag == "local":
            view_sql = """create or replace view dataset as 
            select * 
            from parquet_scan('%s');
            """ % self.parquet_path
        elif self.path_flag == "s3":
            self.cursor.execute("INSTALL httpfs;")
            self.cursor.execute("LOAD httpfs;")
            self.cursor.execute("set s3_endpoint='%s'" % self.endpoint)
            self.cursor.execute("set s3_access_key_id='%s'" % self.access_key_id)
            self.cursor.execute("set s3_secret_access_key='%s'" % self.secret_access_key)
            self.cursor.execute("set s3_url_style = '%s'" % self.url_stype)
            self.cursor.execute("set s3_use_ssl=%s" % self.use_ssl)
            view_sql = """create or replace view dataset as 
            select * 
            from read_parquet('%s');
            """ % self.parquet_path

        self.cursor.execute(view_sql)

    def select(self, select_cols='*', filter_cond='', limit=None, offset=None, samples=None):
        """
        Select data from parquet file
        @param select_cols: columns you want to select
        @param filter_cond: specify filters to apply to the data
        @param limit: restrict the amount of rows fetched
        @param offset: indicate at which position to start reading the values
        @param samples: query on a sample from the base table
        @return: a dataframe of data selected from the parquet
        """
        self.__create_dataset_view()

        query_sql = "select {cols} from dataset".format(cols=select_cols)

        # add where condition
        if filter_cond:
            query_sql = query_sql + ' where ' + filter_cond

        # add limit
        if limit:
            query_sql = query_sql + ' limit ' + str(limit)

        # add offset
        if offset:
            query_sql = query_sql + ' offset ' + str(offset)

        # add random samples
        if samples:
            # to do: implement %
            query_sql = 'select * from (' + query_sql + ') using sample %d rows' % samples

        # print(query_sql)
        df = self.cursor.execute(query_sql).fetch_df()
        return df

    def get_metadata(self):
        """
        Get metadata from parquet schema
        @return: 2 dicts of metadata: dsdl meta and statistics
        """
        if self.path_flag == "local":
            if not os.path.exists(self.parquet_path):
                raise Exception("parquet file does not exist")

            meta_dict = pq.read_schema(self.parquet_path)

        elif self.path_flag == "s3":
            meta_dict = pq.read_schema(self.parquet_key, filesystem=self.fs)

        stat_meta = eval(meta_dict.metadata[b'statistics'])
        return stat_meta

    def get_schema(self):
        """
        Get the schema of a parquet
        @return: a parquet schema
        """
        if self.path_flag == "local":
            if not os.path.exists(self.parquet_path):
                raise Exception("parquet file does not exist")
            schema = pq.read_schema(self.parquet_path)

        elif self.path_flag == "s3":
            schema = pq.read_schema(self.parquet_key, filesystem=self.fs)

        return schema

    def query(self, sql):
        self.__create_dataset_view()
        return self.cursor.execute(sql).fetch_df()


class SplitReader(ParquetReader):
    """
    A class to read parquet base on dataset name and split name
    """

    def __init__(self, dataset_name, split_name):
        self.dataset_name = dataset_name
        self.split_name = split_name
        self.db_client = admin.DBClient()
        self.parquet_path = ''

        if self.db_client.is_split_local_exist(dataset_name, split_name):
            self.parquet_path = self.db_client.get_local_split_path(dataset_name, split_name)
        else:
            self.s3_client = ops.OssClient(endpoint_url=endpoint_url, aws_access_key_id=aws_access_key_id,
                                           aws_secret_access_key=aws_secret_access_key, region_name=region_name)
            if self.s3_client.is_split_remote_exist(default_bucket, dataset_name, split_name):
                self.parquet_path = "s3://%s/%s/parquet/%s.parquet" % (default_bucket, dataset_name, split_name)

        if not self.parquet_path:
            print("Can not find the split named %s of dataset %s neither in local nor remote repo" % (
                split_name, dataset_name))
            exit()

        super(SplitReader, self).__init__(self.parquet_path, endpoint_url.replace("http://", ""), aws_access_key_id,
                                          aws_secret_access_key)

    def get_image_samples(self, sample_number):
        image_list = self.select('image', samples=sample_number)['image'].tolist()
        if self.path_flag == "local":
            dataset_path = self.db_client.get_local_dataset_path(self.dataset_name)
            image_list = [os.path.join(dataset_path, img) for img in image_list]
        elif self.path_flag == "s3":
            dataset_path = "s3://%s/%s/" % (default_bucket, self.dataset_name)
            image_list = [dataset_path + img for img in image_list]
        return image_list


class DSDLParquet:
    """
    A class to handle operations of dsdl parquet file
    """

    def __init__(self, dataframe, parquet_path, schema, statistics=None):
        self.parquet_path = parquet_path
        self.meta_dict = schema.metadata
        if statistics:
            self.meta_dict[b'statistics'] = str(statistics)

        self.dataframe = dataframe
        self.schema = schema.with_metadata(self.meta_dict)
        self.table = pa.Table.from_pandas(self.dataframe, schema=self.schema, preserve_index=False)

    def save(self):
        """
        Save the dsdl parquet to a specified path
        @return:
        """
        pq.write_table(self.table, self.parquet_path)


class Split:
    """
    A class to handle a split
    """

    def __init__(self, dataset_name, split_name, storage_path=''):
        self.db_client = admin.DBClient()
        self.dataset_name = dataset_name
        self.split_name = split_name

        if not self.db_client.is_dataset_local_exist(dataset_name):
            print('There is no local dataset named %s, create local path...' % dataset_name)
            if not storage_path:
                print("no storage path given...")
                exit()
            else:
                self.dataset_path = os.path.join(storage_path, dataset_name)
                self.parquet_folder_path = os.path.join(self.dataset_path, 'parquet')
                self.media_folder_path = os.path.join(self.dataset_path, 'media')
                for p in [self.dataset_path, self.parquet_folder_path, self.media_folder_path]:
                    if not os.path.exists(p):
                        os.mkdir(p)
                self.dataset_info_path = os.path.join(self.parquet_folder_path, 'dataset.yaml')
                self.s3_client = ops.OssClient(endpoint_url=endpoint_url, aws_access_key_id=aws_access_key_id,
                                               aws_secret_access_key=aws_secret_access_key, region_name=region_name)
                self.s3_client.download_file(default_bucket, dataset_name + '/parquet/dataset.yaml',
                                             self.dataset_info_path)

        else:
            self.dataset_path = self.db_client.get_local_dataset_path(self.dataset_name)
            self.parquet_folder_path = os.path.join(self.dataset_path, 'parquet')
            self.media_folder_path = os.path.join(self.dataset_path, 'media')

        self.parquet_path = os.path.join(self.parquet_folder_path, self.split_name + '.parquet')

    def is_local_exist(self):
        """
        Check whether the split is locally existed
        @return:
        """
        return self.db_client.is_split_local_exist(self.dataset_name, self.split_name)

    def save(self, dataframe, schema, type, label, media, statistics):
        """
        Save the split
        @param dataframe: data in dataframe format
        @param schema:
        @param dsdl_meta:
        @param statistics:
        @return:
        """
        media_num = statistics['split_stat']['media_num']
        media_size = statistics['split_stat']['media_size']
        dsdl_parquet = DSDLParquet(dataframe, self.parquet_path, schema, statistics)
        dsdl_parquet.save()
        self.db_client.register_split(self.dataset_name, self.split_name, type, label, media, media_num, media_size)


if __name__ == '__main__':
    db_client = admin.DBClient()
    path = db_client.get_local_split_path('CIFAR-10', 'train')
    split_reader = SplitReader('CIFAR-10-Auto', 'train')

    # print(split_reader.select(filter_cond="label='bird'", select_cols="image,label", samples=1500,
    #                           limit=800, offset=100))
    # dsdl_meta, stat_meta = split_reader.get_metadata()
    # print(dsdl_meta)
    # print(stat_meta)
    # #
    # df = split_reader.select(filter_cond="label='bird'", select_cols="image,label", samples=1500,
    #                          limit=800, offset=100)
    # schema = split_reader.get_schema()
    # meta = split_reader.get_metadata()
    # split = DSDLParquet(df, 'D:\\DSDL_STORE\\test2.parquet', schema, statistics={'test': 'test'})
    # split.save()
    #
    # parquet_reader = ParquetReader('D:\\DSDL_STORE\\STL-10\\STL-10\\parquet\\train.parquet')
    # dsdl_meta, stat_meta = parquet_reader.get_metadata()
    # print(dsdl_meta)
    # print(stat_meta)
    # meta_dict = pq.read_schema(db_client.get_local_split_path('CIFAR-100', 'train'))
    print(split_reader.get_image_samples(10))
    s3_path = "s3://dsdldata/CIFAR-10/parquet/test.parquet"
    parquet_reader = ParquetReader(s3_path, '10.140.0.94:9800', 'ailabminio', '123123123')
    df = parquet_reader.select(limit=20)
    print(df)
    stat = parquet_reader.get_metadata()
    print(stat)
    schema = parquet_reader.get_schema()
    print(schema)
    # test_df = parquet_reader.query("select * from dataset limit 10")
    # print(test_df)



import os
import logging
from minio import Minio
import time

bucket_name = 'dsdldata'


class OSS_OPS:
    def __init__(self, aws_access_key_id, aws_secret_access_key, endpoint_url, region_name):
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.endpoint_url = endpoint_url
        self.region_name = region_name

        self.s3_client = Minio(endpoint=self.endpoint_url, access_key=self.aws_access_key_id,
                               secret_key=self.aws_secret_access_key, region=self.region_name, secure=False)

    def download_file(self, bucket, remote_file, local_file):
        """
        download a file from s3
        :param bucket: bucket name
        :param remote_file: remote file name
        :param local_file: local file name
        :return:
        """
        try:
            self.s3_client.fget_object(bucket, remote_file, local_file)
        except Exception as e:
            logging.error(e)
            return False
        return True

    def list_objects(self, bucket, remote_directory='', recursive=False):
        """
        list objects in a bucket
        :param bucket:
        :return:
        """
        obj_list = self.s3_client.list_objects(
            bucket_name=bucket,
            prefix=remote_directory,
            recursive=recursive,
        )
        obj_list = list(obj_list)
        # print("total file number: %d" % len(obj_list))
        return obj_list

    def get_all_dir(self, dirs, bucket, remote_directory):
        dirs.append(remote_directory)
        child_list = self.list_objects(bucket=bucket, remote_directory=remote_directory)
        child_dir_list = [x for x in child_list if x.is_dir]
        if len(child_dir_list) != 0:
            for child_dir in child_dir_list:
                self.get_all_dir(dirs, bucket, child_dir.object_name)

    def print_progress(self, iteration, total, start_time, prefix='', suffix='', decimals=1, bar_length=100):
        """
        Call in a loop to create a terminal progress bar
        @params:
            iteration   - Required  : current iteration (Int)
            total       - Required  : total iterations (Int)
            prefix      - Optional  : prefix string (Str)
            suffix      - Optional  : suffix string (Str)
            decimals    - Optional  : positive number of decimals in percent complete (Int)
            barLength   - Optional  : character length of bar (Int)
        """
        import sys
        format_str = "{0:." + str(decimals) + "f}"
        percent = format_str.format(100 * (iteration / float(total)))
        filled_length = int(round(bar_length * iteration / float(total)))
        bar = '#' * filled_length + '-' * (bar_length - filled_length)
        dur = time.perf_counter() - start_time
        sys.stdout.write('\r%s |%s| %ds %s%s %s' % (prefix, bar, dur, percent, '%', suffix)),
        if iteration == total:
            sys.stdout.write('\n')
        sys.stdout.flush()

    def download_directory(self, bucket, remote_directory, local_directory):
        """
        download a directory from s3
        :param bucket: bucket name
        :param remote_directory: remote directory name
        :param local_directory: local directory name
        :return:
        """
        print("check local path...")
        dirs_list = []
        self.get_all_dir(dirs=dirs_list, bucket=bucket, remote_directory=remote_directory)
        for d in dirs_list:
            path = os.path.join(local_directory, d)
            if not os.path.exists(path):
                os.mkdir(path)

        print("start download...")
        try:
            file_list = self.list_objects(bucket, remote_directory, recursive=True)
            process = 0
            file_number = len(file_list)
            start_time = time.perf_counter()

            self.print_progress(process, file_number, start_time, prefix='Downloading', suffix='Complete',
                                bar_length=50)

            for file in file_list:
                local_file = os.path.join(local_directory, file.object_name)
                self.download_file(bucket, file.object_name, local_file)
                process += 1
                self.print_progress(process, file_number, start_time, prefix='Download', suffix='Complete',
                                    bar_length=50)
        except Exception as e:
            logging.error(e)
            return False
        print('Download Complete')
        return True


if __name__ == '__main__':
    aws_access_key_id = "ailabminio"
    aws_secret_access_key = "123123123"
    endpoint_url = "10.140.0.94:9800"
    region_name = "ailab"
    minio_client = OSS_OPS(endpoint_url=endpoint_url, aws_access_key_id=aws_access_key_id,
                           aws_secret_access_key=aws_secret_access_key, region_name=region_name)
    # objs = minio_client.list_objects('dsdldata', 'CIFAR-10/parquet')
    # for obj in objs:
    #     print(obj.object_name, obj.is_dir)

    dir_list = []
    minio_client.get_all_dir(dir_list, 'testdata', 'test_data/')
    print(dir_list)

    # for d in dir_list:
    #     path = os.path.join('D:\\', d)
    #     print(path)
    #     if not os.path.exists(path):
    #         os.mkdir(path)

    # minio_client.download_file('dsdldata', 'CIFAR-10/media/000000042381.png', 'D:\DSDL_DATA\CIFAR-10/media/000000042381.png')
    # minio_client.download_directory('dsdldata', 'CIFAR-10/', 'D:\\DSDL_DATA')



from commands.cmdbase import CmdBase
from commands.const import DSDL_CLI_DATASET_NAME
from commons.argument_parser import EnvDefaultVar
from rich import print as rprint
from rich.pretty import pprint
from tabulate import tabulate
from utils import admin


class Ls(CmdBase):
    """Show all the available datasets and its storage path.

    Args:
        CmdBase (_type_): _description_
    """
    def init_parser(self, subparsers):
        """_summary_

        Args:
            subparsers (_type_): _description_
        """
        
        status_parser = subparsers.add_parser('ls', help = 'show local datasets', example = 'ls.example')
        status_parser.add_argument('-s',
                                   '--skip-header', 
                                   help = 'toggle display for headers',
                                   action = 'store_true',
                                   required = False)
        status_parser.add_argument('-v',
                                   '--verbose',
                                   action = 'count',
                                   default = 0,
                                   help = 'display detailed dataset info',
                                   required = False)
        return status_parser
    
    def get_dataset_df(self):
        db_client = admin.DBClient()
        dataset_list = db_client.get_sqlite_dict_list('select * from dataset')
        datasplit_list = db_client.get_sqlite_dict_list('select * from split')
        return dataset_list, datasplit_list
    
    def datalist_trim(self, dataset_list, datasplit_list):
        dataset_list_trim = []
        datasplit_list_trim = []
        dataset_verbose_list = ['created_time','updated_time','dataset_path']
        datasplit_verbose_list = ['created_time','updated_time']
        for idx, elem in enumerate(dataset_list):
            dataset_list_trim.append({key:dataset_list[idx][key] for key in dataset_list[idx] if key not in dataset_verbose_list})
            
        for idx, elem in enumerate(datasplit_list):
            datasplit_list_trim.append({key:datasplit_list[idx][key] for key in datasplit_list[idx] if key not in datasplit_verbose_list})
        
        return dataset_list_trim, datasplit_list_trim
    
    def cmd_entry(self, cmdargs, config):
        """
        Entry point for the command.

        Args:
            self:
            args:
            config:

        Returns:

        """
        
        dataset_list, datasplit_list = self.get_dataset_df()
        dataset_list_trim, datasplit_list_trim = self.datalist_trim(dataset_list, datasplit_list)

        # default display no argument is given
        if not (cmdargs.skip_header or cmdargs.verbose):
            # rprint(tabulate(datasplit_list_trim, headers='keys', tablefmt='plain'))
            rprint(tabulate(dataset_list_trim, headers='keys', tablefmt='plain'))
            
        if cmdargs.verbose:
            if cmdargs.verbose >= 1:
                # verbose and skip head
                if cmdargs.skip_header:
                    rprint(tabulate(dataset_list, tablefmt='plain'))
                # verbose but not skip head
                else:
                    rprint(tabulate(dataset_list, headers = 'keys', tablefmt='plain'))
                    
        if cmdargs.skip_header:
            # not verbose but skip head
            if not cmdargs.verbose:
                rprint(tabulate(dataset_list_trim, tablefmt='plain'))


import os
import sys
import sqlite3
from typing import List

sys.path.append(f"{os.path.dirname(os.path.abspath(__file__))}/../../../")

import pandas as pd
import streamlit as st
from stqdm import stqdm

from src.constants.colors import COLOR_MAP
from src.utils.display import load_and_annotate_image
import utils.admin as admin


st.set_page_config(
    # Change the title of the web page
    "Explore dataset",
    # Change the icon of the web page to a magnifier glass
    page_icon=":mag_right:",
)

# download the dataset to specified location and unzip it


conn = sqlite3.connect(admin.DB_PATH)

# show all records of dataset which already stored in local
print(conn.execute("select * from dataset").fetchall())

dbclient = admin.DBClient()

dbclient.register_dataset(
    storage_name="default",
    dataset_name="coco-demo",
    dataset_path="/root/.dsdl/datasets/coco_demo/data/annotations",
    label=1,
    media=1,
    media_num=300,
    media_size=1000,
)
# fetch the newly inserted record of this dataset
dbclient.get_local_dataset_path("coco-demo")

data_dir_path = dbclient.get_local_dataset_path("coco-demo")
train_data_file_path = os.path.join(data_dir_path, "coco_train_2020.parquet.gzip")
val_data_file_path = os.path.join(data_dir_path, "coco_val_2020.parquet.gzip")


def load_all_annotations() -> pd.DataFrame:
    return pd.read_parquet(train_data_file_path)
    # return pd.read_parquet(val_data_file_path)


# Cache the result to avoid reloading the full dataset
# and recomputing everytime we run the app
@st.experimental_memo()
def get_category_count() -> pd.DataFrame:
    return load_all_annotations().category_name.value_counts()


with st.sidebar:
    # Display a summary of the labels in the dataset
    category_count = get_category_count()
    st.write(category_count)
    st.write(f"{len(category_count)} Labels available")

    # Add a multiselect input to selection multiple labels
    available_labels = category_count.index.tolist()
    labels_to_display = st.multiselect(
        "Labels to display", available_labels, default=["toaster"]
    )


# Cache the last three requests
@st.experimental_memo(max_entries=3)
def get_selected_images(selected_categories: List[str]) -> pd.DataFrame:
    # Return all the images containing at least
    # one of the selected categories
    # return all the annotations on the selected images
    all_annotations = load_all_annotations()
    return all_annotations.groupby("image_name").filter(
        lambda g: g.category_name.isin(selected_categories).any()
    )


annotations = get_selected_images(labels_to_display)
n_images = len(annotations[["image_name", "coco_url"]].drop_duplicates())
with st.sidebar:
    st.write(f"{n_images} images selected")

# Loop on images and display them
for (image_name, coco_url), image_annotations in stqdm(
    annotations.groupby(["image_name", "coco_url"]),
    desc="Displaying Images",
    total=n_images,
):
    st.image(
        load_and_annotate_image(coco_url, image_annotations, color_map=COLOR_MAP),
        caption=image_name,
    )



# import the Package
import os
import sys
import base64
from io import BytesIO
import urllib

sys.path.append(f'{os.path.dirname(os.path.abspath(__file__))}/../../../')

import duckdb
import streamlit as st
from PIL import Image
import pandas as pd

from utils.admin import DBClient

class_name = [
    "airplane",
    "automobile",
    "bird",
    "cat",
    "deer",
    "dog",
    "frog",
    "horse",
    "ship",
    "truck",
]


def main():
    st.title("CIFAR-10")
    st.sidebar.title("CIFAR-10")
    app_mode = st.sidebar.selectbox(
        "Choose the app mode",
        ["Show dataset introductions", "Show dataset details", "Show source code", "Explore the app"],
    )
    if app_mode == "Show dataset introductions":
        st.sidebar.success("Displaying the introductions of dataset.")
        st.write(INTRO_DOC_STR)
    elif app_mode == "Show dataset details":
        st.sidebar.success("Displaying the details of dataset.")
        st.json(DATASET_DOC_STR, expanded=True)
    elif app_mode == "Show source code":
        st.sidebar.success("Displaying the dataset source code.")
        st.code(CODE_FILE_STR, language="python")
    elif app_mode == "Explore the app":
        st.sidebar.success("Exploring the dataset.")
        explore_app()


def parquet_filter(parquet_path, select_cols="*", filter_cond="", limit=None, offset=None, samples=None):
    cursor = duckdb.connect(database=":memory:")
    view_sql = """create or replace view dataset as select * from parquet_scan('%s');""" % parquet_path
    cursor.execute(view_sql)
    query_sql = "select {cols} from dataset".format(cols=select_cols)

    # add where condition
    if filter_cond:
        query_sql = query_sql + " where " + filter_cond

    # add limit
    if limit:
        query_sql = query_sql + " limit " + str(limit)

    # add offset
    if offset:
        query_sql = query_sql + " offset " + str(offset)

    # add random samples
    if samples:
        # to do: implement %
        query_sql = "select * from (" + query_sql + ") using sample %d rows" % samples

    # print(query_sql)
    df = cursor.execute(query_sql).fetch_df()
    return df


def explore_app():
    # the following varibles should be read from dsdl database by duckdb
    dbcli = DBClient()
    dataset_dir_path = dbcli.get_local_dataset_path('CIFAR-10')
    parquet_dir_path = os.path.join(dataset_dir_path, 'parquet')
    parquet_train_file = os.path.join(parquet_dir_path, 'train.parquet')
    parquet_test_file = os.path.join(parquet_dir_path, 'test.parquet')

    TRAIN_DF = parquet_filter(
        parquet_path=parquet_train_file,
        select_cols="*",
        filter_cond="",
        limit=None,
        offset=None,
        samples=None,
    )
    TEST_DF = parquet_filter(
        parquet_path=parquet_test_file,
        select_cols="*",
        filter_cond="",
        limit=None,
        offset=None,
        samples=None,
    )

    TRAIN_DF['split'] = 'train'
    TEST_DF['split'] = 'test'
    DF = pd.concat([TRAIN_DF, TEST_DF], ignore_index=True)
    # DF['path'] = dataset_dir_path + DF['image']
    DF['path'] = DF['image'].apply(lambda x: os.path.join(dataset_dir_path, x))
    DF['thumbnail'] = DF.path.map(lambda x: image_formatter(x))
    DF = pd.DataFrame(
        {
            'name': DF["image"],
            'label': DF["label"],
            'split': DF["split"],
            'path': DF["path"],
            'thumbnail': DF["thumbnail"],
        }
    )

    DF_HTML = convert_df(DF.head(100))

    pd.set_option('display.max_colwidth', -1)
    st.markdown(DF_HTML, unsafe_allow_html=True)


@st.cache
def convert_df(input_df):
    # IMPORTANT: Cache the conversion to prevent computation on every rerun
    return input_df.to_html(
        formatters={'img': image_formatter},
        escape=False,
    )


def get_thumbnail(path):
    i = Image.open(path)
    i.thumbnail((320, 320), Image.LANCZOS)
    return i


def image_base64(im):
    if isinstance(im, str):
        im = get_thumbnail(im)
    with BytesIO() as buffer:
        im.save(buffer, 'jpeg')
        return base64.b64encode(buffer.getvalue()).decode()


def image_formatter(im):
    return f'<img src="data:image/jpeg;base64,{image_base64(im)}">'


intro_doc_path = "https://huggingface.co/datasets/cifar10/raw/main/README.md"
dataset_doc_path = "https://huggingface.co/datasets/cifar10/raw/main/dataset_infos.json"
code_file_path = "https://huggingface.co/datasets/cifar10/raw/main/cifar10.py"

# Download a single file and make its content available as a string.
@st.experimental_singleton(show_spinner=False)
def get_web_file_content_as_string(url):
    """
    Get raw data content from a web file.

    :param  type: local or web
    :returns file content in string format
    """
    response = urllib.request.urlopen(url)
    return response.read().decode("utf-8")


INTRO_DOC_STR = get_web_file_content_as_string(intro_doc_path)
DATASET_DOC_STR = get_web_file_content_as_string(dataset_doc_path)
CODE_FILE_STR = get_web_file_content_as_string(code_file_path)


if __name__ == "__main__":
    main()
